@@ -2,7 +2,7 @@
 #
 # Apache::Session
 # Apache persistent user sessions
-# Copyright(c) 1998, 1999 Jeffrey William Baker (jeffrey@kathyandjeffrey.net)
+# Copyright(c) 1998, 1999, 2000, 2001, 2004 Jeffrey William Baker (jwbaker@acm.org)
 # Distribute under the Artistic License
 #
 #############################################################################
@@ -13,12 +13,12 @@ Apache::Session - A persistence framework for session data
 
 =head1 SYNOPSIS
 
-  use Apache::Session::DBI;
+  use Apache::Session::MySQL;
   
   my %session;
   
   #make a fresh session for a first-time visitor
-  tie %session, 'Apache::Session::DBI';
+  tie %session, 'Apache::Session::MySQL';
 
   #stick some stuff in it
   $session{visa_number} = "1234 5678 9876 5432";
@@ -28,9 +28,9 @@ Apache::Session - A persistence framework for session data
   
   #...time passes...
   
-  #get the session data back out again
+  #get the session data back out again during some other request
   my %session;
-  tie %session, 'Apache::Session::DBI', $id;
+  tie %session, 'Apache::Session::MySQL', $id;
   
   &validate($session{visa_number});
   
@@ -38,37 +38,29 @@ Apache::Session - A persistence framework for session data
   tied(%session)->delete;
   
 
-=head2 NOTE
-
-There was an earlier attempt at creating a session handler, which 
-existed until Apache::Session 0.17.  This version is completely 
-incompatible with that version.  If you are using Embperl, as of this
-writing you need to use Apache::Session 0.17.
-
-=head1 THE GAME
-
-HTTP is a stateless protocol, which makes it difficult to track 
-a user between requests.  Apache::Session bridges this problem.
-
 =head1 DESCRIPTION
 
 Apache::Session is a persistence framework which is particularly useful
 for tracking session data between httpd requests.  Apache::Session is
 designed to work with Apache and mod_perl, but it should work under
-CGI and other web servers.
+CGI and other web servers, and it also works outside of a web server
+altogether.
 
-Apache::Session consists of three components: the interface, the object
-store, and the lock manager.  The interface is defined in Session.pm,
-which is meant to be easily subclassed.  The object store is implemented
-by Session::DBIStore, Session::FileStore, and Session::MemoryStore.  
-Various locking schemes are implemented in Session::PosixFileLocker, 
-Session::SysVSemaphoreLocker, and Session::NullLocker.
+Apache::Session consists of five components: the interface, the object store,
+the lock manager, the ID generator, and the serializer.  The interface is
+defined in Session.pm, which is meant to be easily subclassed.  The object
+store can be the filesystem, a Berkeley DB, a MySQL DB, an Oracle DB, a
+Postgres DB, Sybase, or Informix. Locking is done by lock files, semaphores, or
+the locking capabilities of the various databases.  Serialization is done via
+Storable, and optionally ASCII-fied via MIME or pack().  ID numbers are
+generated via MD5.  The reader is encouraged to extend these capabilities to
+meet his own requirements.
 
 A derived class of Apache::Session is used to tie together the three
-components.  The derived class inherits the interface from Apache::Session,
-and specifies which store and locker classes to use.  Apache::Session::DBI,
-for instance, uses the DBIStore class and the SysVSemaphoreLocker class.
-You can easily plug in your own object store or locker class.
+components.  The derived class inherits the interface from Apache::Session, and
+specifies which store and locker classes to use.  Apache::Session::MySQL, for
+instance, uses the MySQL storage class and also the MySQL locking class. You
+can easily plug in your own object store or locker class.
 
 =head1 INTERFACE
 
@@ -82,16 +74,16 @@ of options that will be passed to the object store and locker classes.
 
 Get a new session using DBI:
 
- tie %session, 'Apache::Session::DBI', undef,
-    { DataSource => 'dbi:Oracle:db' };
+ tie %session, 'Apache::Session::MySQL', undef,
+    { DataSource => 'dbi:mysql:sessions' };
     
 Restore an old session from the database:
 
- tie %session, 'Apache::Session::DBI', $session_id,
-    { DataSource => 'dbi:Oracle:db' };
+ tie %session, 'Apache::Session::MySQL', $session_id,
+    { DataSource => 'dbi:mysql:sessions' };
 
 
-=head2 Storing and retrieving from the session
+=head2 Storing and retrieving data to and from the session
 
 Hey, how much easier could it get?
 
@@ -107,7 +99,6 @@ future use.
 
  my $id = $session{_session_id};
 
-
 =head2 Permanently removing the session from storage
 
  tied(%session)->delete;
@@ -126,7 +117,12 @@ As you put data into the session hash, Session squirrels it away for
 later use.  When you untie() the session hash, or it passes out of
 scope, Session checks to see if anything has changed. If so, Session 
 gains an exclusive lock and writes the session to the data store.  
-It then releases any locks it has acquired.
+It then releases any locks it has acquired.  
+
+Note that Apache::Session does only a shallow check to see if anything has
+changed.  If nothing changes in the top level tied hash, the data will not be
+updated in the backing store.  You are encouraged to timestamp the session hash
+so that it is sure to be updated.
 
 When you call the delete() method on the session object, the
 object is immediately removed from the object store, if possible.
@@ -134,33 +130,31 @@ object is immediately removed from the object store, if possible.
 When Session encounters an error, it calls die().  You will probably 
 want to wrap your session logic in an eval block to trap these errors.
 
+=head1 LOCKING AND TRANSACTIONS
+
+By default, most Apache::Session implementations only do locking to prevent
+data corruption.  The locking scheme does not provide transactional
+consistency, such as you might get from a relational database.  If you desire
+transactional consistency, you must provide the Transaction argument with a
+true value when you tie the session hash.  For example:
+
+ tie %s, 'Apache::Session::File', $id {
+    Directory     => '/tmp/sessions',
+    LockDirectory => '/var/lock/sessions',
+    Transaction   => 1
+ };
+
+Note that the Transaction argument has no practical effect on the MySQL and
+Postgres implementations.  The MySQL implementation only supports exclusive
+locking, and the Postgres implementation uses the transaction features of that
+database.
+
 =head1 IMPLEMENTATION
 
 The way you implement Apache::Session depends on what you are
 trying to accomplish.  Here are some hints on which classes to
 use in what situations
 
-=over 4
-
-=item Single machine *nix Apache
-
-Use DBIStore and SysVSemaphoreLocker
-
-=item Single machine Win32 Apache
-
-Use DBIStore or MemoryStore, if persistence between server invocations is not
-neccessary.  Use the NullLocker for best speed.
-
-=item Multiple *nix machines
-
-Use DBIStore and the DaemonLocker, or use PosixFileLocker on an NFS mount
-
-=item Multiple machines on multiple platforms
-
-Use DBIStore and DaemonLocker
-
-=back
-
 =head1 STRATEGIES
 
 Apache::Session is mainly designed to track user session between 
@@ -213,7 +207,10 @@ is a rather religious topic among Apache users.  This example uses cookies.
 The implementation of a path info system is left as an exercise for the
 reader.
 
- use Apache::Session::DBI;
+Note that Apache::Session::Generate::ModUsertrack uses Apache's mod_usertrack
+cookies to generate and maintain session IDs.
+
+ use Apache::Session::MySQL;
  use Apache;
 
  use strict;
@@ -228,11 +225,14 @@ reader.
  #or a new session if we got no cookie
 
  my %session;
- tie %session, 'Apache::Session::DBI', $cookie,
-     {DataSource => 'dbi:mysql:sessions', #these arguments are
+ tie %session, 'Apache::Session::MySQL', $cookie, {
+      DataSource => 'dbi:mysql:sessions', #these arguments are
       UserName   => 'mySQL_user',         #required when using
-      Password   => 'password'            #DBIStore.pm
-     };
+      Password   => 'password',           #MySQL.pm
+      LockDataSource => 'dbi:mysql:sessions',
+      LockUserName   => 'mySQL_user',
+      LockPassword   => 'password'
+ };
 
  #Might be a new session, so lets give them their cookie back
 
@@ -243,25 +243,36 @@ reader.
 
 =head1 SEE ALSO
 
-Apache::Session::DBIStore, Apache::Session::FileStore, 
-Apache::Session::MemoryStore, Apache::Session::PosixFileLocker,
-Apache::Session::SysVSemaphoreLocker, Apache::Session::NullLocker
-Apache::Session::TreeStore, Apache::Session::Counted
+Apache::Session::MySQL, Apache::Session::Postgres, Apache::Session::File,
+Apache::Session::DB_File, Apache::Session::Oracle, Apache::Session::Sybase
 
-The O Reilly book "Apache Modules in Perl and C" has a chapter
-on keeping state.
+The O Reilly book "Apache Modules in Perl and C", by Doug MacEachern and
+Lincoln Stein, has a chapter on keeping state.
 
 =head1 AUTHORS
 
-Jeffrey Baker <jeffrey@kathyandjeffrey.net> is the author of 
+Jeffrey Baker <jwbaker@acm.org> is the author of 
 Apache::Session.
 
+Tatsuhiko Miyagawa <miyagawa@bulknews.net> is the author of 
+Generate::ModUniqueID and Generate::ModUsertrack
+
+Erik Rantapaa <rantapaa@fanbuzz.com> found errors in both Lock::File
+and Store::File
+
+Bart Schaefer <schaefer@zanshin.com> notified me of a bug in 
+Lock::File.
+
+Chris Winters <cwinters@intes.net> contributed the Sybase code.
+
+Michael Schout <mschout@gkg.net> fixed a commit policy bug in 1.51.
+
 Andreas J. Koenig <andreas.koenig@anima.de> contributed valuable CPAN
 advice and also Apache::Session::Tree and Apache::Session::Counted.
 
 Gerald Richter <richter@ecos.de> had the idea for a tied hash interface
 and provided the initial code for it.  He also uses Apache::Session in
-his Embperl module.
+his Embperl module and is the author of Apache::Session::Embperl
 
 Jochen Wiedmann <joe@ipsoft.de> contributed patches for bugs and
 improved performance.
@@ -279,6 +290,21 @@ storage-independent object interface with input from:
   Jules Bean <jmlb2@hermes.cam.ac.uk>
   Lincoln Stein <lstein@cshl.org>
 
+Jamie LeTaul <jletual@kmtechnologies.com> fixed file locking on Windows.
+
+Scott McWhirter <scott@surreytech.co.uk> contributed verbose error messages for
+file locking.
+
+Corris Randall <corris@line6.net> gave us the option to use any table name in
+the MySQL store.
+
+Oliver Maul <oliver.maul@ixos.de> updated the Sybase modules
+
+Innumerable users sent a patch for the reversed file age test in the file
+locking module.
+
+Langen Mike <mike.langen@tamedia.ch> contributed Informix modules.
+
 =cut
 
 package Apache::Session;
@@ -286,9 +312,7 @@ package Apache::Session;
 use strict;
 use vars qw($VERSION);
 
-$VERSION = '1.00';
-
-use MD5; #yes, you need MD5.pm
+$VERSION = '1.70_01';
 
 #State constants
 #
@@ -344,34 +368,47 @@ sub TIEHASH {
     my $session_id = shift;
     my $args       = shift || {};
 
-    if(ref $args ne "HASH") {
-        die "Additional arguments should be in the form of a hash reference";
-    }
-
     #Set-up the data structure and make it an object
     #of our class
     
     my $self = {
         args         => $args,
         data         => { _session_id => $session_id },
+        serialized   => undef,
         lock         => 0,
-        lock_manager => undef,
+        status       => 0,
+        lock_manager => undef,  # These two are object refs ...
         object_store => undef,
-        status       => 0
+        generate     => undef,  # but these three are subroutine refs
+        serialize    => undef,
+        unserialize  => undef,
     };
     
     bless $self, $class;
 
+    $self->populate;
+
+
     #If a session ID was passed in, this is an old hash.
     #If not, it is a fresh one.
 
-    if (defined $session_id) {
+    if (defined $session_id  && $session_id) {
+        
+        #check the session ID for remote exploitation attempts
+        #this will die() on suspicious session IDs.        
+
+        &{$self->{validate}}($self);        
+        
+        if (exists $args->{Transaction} && $args->{Transaction}) {
+            $self->acquire_write_lock;
+        }
+        
         $self->{status} &= ($self->{status} ^ NEW);
         $self->restore;
     }
     else {
         $self->{status} |= NEW;
-        $self->{data}->{_session_id} = generate_id();
+        &{$self->{generate}}($self);
         $self->save;
     }
     
@@ -456,11 +493,8 @@ sub restore {
     
     $self->acquire_read_lock;
 
-    if (!defined $self->{object_store}) {
-        $self->{object_store} = $self->get_object_store;
-    }
-    
     $self->{object_store}->materialize($self);
+    &{$self->{unserialize}}($self);
     
     $self->{status} &= ($self->{status} ^ MODIFIED);
     $self->{status} |= SYNCED
@@ -476,11 +510,7 @@ sub save {
     );
     
     $self->acquire_write_lock;
-    
-    if (!defined $self->{object_store}) {
-        $self->{object_store} = $self->get_object_store;
-    }
-    
+
     if ($self->{status} & DELETED) {
         $self->{object_store}->remove($self);
         $self->{status} |= SYNCED;
@@ -489,12 +519,14 @@ sub save {
         return;
     }
     if ($self->{status} & MODIFIED) {
+        &{$self->{serialize}}($self);
         $self->{object_store}->update($self);
         $self->{status} &= ($self->{status} ^ MODIFIED);
         $self->{status} |= SYNCED;
         return;
     }
     if ($self->{status} & NEW) {
+        &{$self->{serialize}}($self);
         $self->{object_store}->insert($self);
         $self->{status} &= ($self->{status} ^ NEW);
         $self->{status} |= SYNCED;
@@ -539,10 +571,6 @@ sub acquire_read_lock  {
 
     return if ($self->{lock} & READ_LOCK);
 
-    if (!defined $self->{lock_manager}) {
-        $self->{lock_manager} = $self->get_lock_manager;
-    }
-
     $self->{lock_manager}->acquire_read_lock($self);
 
     $self->{lock} |= READ_LOCK;
@@ -553,10 +581,6 @@ sub acquire_write_lock {
 
     return if ($self->{lock} & WRITE_LOCK);
 
-    if (!defined $self->{lock_manager}) {
-        $self->{lock_manager} = $self->get_lock_manager;
-    }
-
     $self->{lock_manager}->acquire_write_lock($self);
 
     $self->{lock} |= WRITE_LOCK;
@@ -567,10 +591,6 @@ sub release_read_lock {
 
     return unless ($self->{lock} & READ_LOCK);
 
-    if (!defined $self->{lock_manager}) {
-        $self->{lock_manager} = $self->get_lock_manager;
-    }
-
     $self->{lock_manager}->release_read_lock($self);
 
     $self->{lock} &= ($self->{lock} ^ READ_LOCK);
@@ -581,10 +601,6 @@ sub release_write_lock {
 
     return unless ($self->{lock} & WRITE_LOCK);
 
-    if (!defined $self->{lock_manager}) {
-        $self->{lock_manager} = $self->get_lock_manager;
-    }
-
     $self->{lock_manager}->release_write_lock($self);
     
     $self->{lock} &= ($self->{lock} ^ WRITE_LOCK);
@@ -595,26 +611,10 @@ sub release_all_locks {
     
     return unless ($self->{lock} & READ_LOCK || $self->{lock} & WRITE_LOCK);
     
-    if (!defined $self->{lock_manager}) {
-        $self->{lock_manager} = $self->get_lock_manager;
-    }
-
     $self->{lock_manager}->release_all_locks($self);
 
     $self->{lock} &= ($self->{lock} ^ READ_LOCK);
     $self->{lock} &= ($self->{lock} ^ WRITE_LOCK);
 }        
 
-
-
-#
-#Utility methods
-#
-
-
-
-sub generate_id {
-    return substr(MD5->hexhash(time(). {}. rand(). $$. 'blah'), 0, 16);
-}
-
 1;
